@@ -1,7 +1,9 @@
 //! ASN.1 `BOOLEAN` support.
 
-use crate::{asn1::Any, Encodable, Encoder, Error, Header, Length, Result, Tag, Tagged};
-use core::convert::TryFrom;
+use crate::{
+    asn1::Any, ByteSlice, DecodeValue, Decoder, EncodeValue, Encoder, Error, ErrorKind, FixedTag,
+    Length, OrdIsValueOrd, Result, Tag,
+};
 
 /// Byte used to encode `true` in ASN.1 DER. From X.690 Section 11.1:
 ///
@@ -12,36 +14,55 @@ const TRUE_OCTET: u8 = 0b11111111;
 /// Byte used to encode `false` in ASN.1 DER.
 const FALSE_OCTET: u8 = 0b00000000;
 
-impl TryFrom<Any<'_>> for bool {
-    type Error = Error;
+impl<'a> DecodeValue<'a> for bool {
+    fn decode_value(decoder: &mut Decoder<'a>, length: Length) -> Result<Self> {
+        if length != Length::ONE {
+            return Err(decoder.error(ErrorKind::Length { tag: Self::TAG }));
+        }
 
-    fn try_from(any: Any<'_>) -> Result<bool> {
-        let tag = any.tag().assert_eq(Tag::Boolean)?;
-
-        match any.as_bytes() {
-            [FALSE_OCTET] => Ok(false),
-            [TRUE_OCTET] => Ok(true),
-            _ => Err(tag.non_canonical_error()),
+        match decoder.byte()? {
+            FALSE_OCTET => Ok(false),
+            TRUE_OCTET => Ok(true),
+            _ => Err(Self::TAG.non_canonical_error()),
         }
     }
 }
 
-impl Encodable for bool {
-    fn encoded_len(&self) -> Result<Length> {
-        Length::ONE.for_tlv()
+impl EncodeValue for bool {
+    fn value_len(&self) -> Result<Length> {
+        Ok(Length::ONE)
     }
 
-    fn encode(&self, encoder: &mut Encoder<'_>) -> Result<()> {
-        Header::new(Self::TAG, Length::ONE)?.encode(encoder)?;
-        let byte = if *self { TRUE_OCTET } else { FALSE_OCTET };
-        encoder.byte(byte)
+    fn encode_value(&self, encoder: &mut Encoder<'_>) -> Result<()> {
+        encoder.byte(if *self { TRUE_OCTET } else { FALSE_OCTET })
     }
 }
 
-impl Tagged for bool {
+impl FixedTag for bool {
     const TAG: Tag = Tag::Boolean;
 }
 
+impl OrdIsValueOrd for bool {}
+
+impl From<bool> for Any<'static> {
+    fn from(value: bool) -> Any<'static> {
+        let value = ByteSlice::from(match value {
+            false => &[FALSE_OCTET],
+            true => &[TRUE_OCTET],
+        });
+
+        Any::from_tag_and_value(Tag::Boolean, value)
+    }
+}
+
+impl TryFrom<Any<'_>> for bool {
+    type Error = Error;
+
+    fn try_from(any: Any<'_>) -> Result<bool> {
+        any.try_into()
+    }
+}
+
 #[cfg(test)]
 mod tests {
     use crate::{Decodable, Encodable};
