@@ -1,9 +1,10 @@
 //! ASN.1 `IA5String` support.
 
 use crate::{
-    asn1::Any, str_slice::StrSlice, Encodable, Encoder, Error, Length, Result, Tag, Tagged,
+    asn1::Any, ByteSlice, DecodeValue, Decoder, EncodeValue, Encoder, Error, FixedTag, Length,
+    OrdIsValueOrd, Result, StrSlice, Tag,
 };
-use core::{convert::TryFrom, fmt, str};
+use core::{fmt, str};
 
 /// ASN.1 `IA5String` type.
 ///
@@ -72,6 +73,28 @@ impl AsRef<[u8]> for Ia5String<'_> {
     }
 }
 
+impl<'a> DecodeValue<'a> for Ia5String<'a> {
+    fn decode_value(decoder: &mut Decoder<'a>, length: Length) -> Result<Self> {
+        Self::new(ByteSlice::decode_value(decoder, length)?.as_bytes())
+    }
+}
+
+impl EncodeValue for Ia5String<'_> {
+    fn value_len(&self) -> Result<Length> {
+        self.inner.value_len()
+    }
+
+    fn encode_value(&self, encoder: &mut Encoder<'_>) -> Result<()> {
+        self.inner.encode_value(encoder)
+    }
+}
+
+impl<'a> FixedTag for Ia5String<'a> {
+    const TAG: Tag = Tag::Ia5String;
+}
+
+impl OrdIsValueOrd for Ia5String<'_> {}
+
 impl<'a> From<&Ia5String<'a>> for Ia5String<'a> {
     fn from(value: &Ia5String<'a>) -> Ia5String<'a> {
         *value
@@ -82,8 +105,7 @@ impl<'a> TryFrom<Any<'a>> for Ia5String<'a> {
     type Error = Error;
 
     fn try_from(any: Any<'a>) -> Result<Ia5String<'a>> {
-        any.tag().assert_eq(Tag::Ia5String)?;
-        Self::new(any.as_bytes())
+        any.decode_into()
     }
 }
 
@@ -99,20 +121,6 @@ impl<'a> From<Ia5String<'a>> for &'a [u8] {
     }
 }
 
-impl<'a> Encodable for Ia5String<'a> {
-    fn encoded_len(&self) -> Result<Length> {
-        Any::from(*self).encoded_len()
-    }
-
-    fn encode(&self, encoder: &mut Encoder<'_>) -> Result<()> {
-        Any::from(*self).encode(encoder)
-    }
-}
-
-impl<'a> Tagged for Ia5String<'a> {
-    const TAG: Tag = Tag::Ia5String;
-}
-
 impl<'a> fmt::Display for Ia5String<'a> {
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         f.write_str(self.as_str())
