@@ -1,18 +1,34 @@
 //! ASN.1 `OBJECT IDENTIFIER`
 
-use crate::{asn1::Any, Encodable, Encoder, Error, Length, Result, Tag, Tagged};
+use crate::{
+    asn1::Any, ByteSlice, DecodeValue, Decoder, EncodeValue, Encoder, Error, FixedTag, Length,
+    OrdIsValueOrd, Result, Tag, Tagged,
+};
 use const_oid::ObjectIdentifier;
-use core::convert::{TryFrom, TryInto};
 
-impl TryFrom<Any<'_>> for ObjectIdentifier {
-    type Error = Error;
-
-    fn try_from(any: Any<'_>) -> Result<ObjectIdentifier> {
-        any.tag().assert_eq(Tag::ObjectIdentifier)?;
-        Ok(ObjectIdentifier::from_bytes(any.as_bytes())?)
+impl DecodeValue<'_> for ObjectIdentifier {
+    fn decode_value(decoder: &mut Decoder<'_>, length: Length) -> Result<Self> {
+        let bytes = ByteSlice::decode_value(decoder, length)?.as_bytes();
+        Ok(Self::from_bytes(bytes)?)
     }
 }
 
+impl EncodeValue for ObjectIdentifier {
+    fn value_len(&self) -> Result<Length> {
+        Length::try_from(self.as_bytes().len())
+    }
+
+    fn encode_value(&self, encoder: &mut Encoder<'_>) -> Result<()> {
+        encoder.bytes(self.as_bytes())
+    }
+}
+
+impl FixedTag for ObjectIdentifier {
+    const TAG: Tag = Tag::ObjectIdentifier;
+}
+
+impl OrdIsValueOrd for ObjectIdentifier {}
+
 impl<'a> From<&'a ObjectIdentifier> for Any<'a> {
     fn from(oid: &'a ObjectIdentifier) -> Any<'a> {
         // Note: ensuring an infallible conversion is possible relies on the
@@ -28,25 +44,19 @@ impl<'a> From<&'a ObjectIdentifier> for Any<'a> {
     }
 }
 
-impl Encodable for ObjectIdentifier {
-    fn encoded_len(&self) -> Result<Length> {
-        Any::from(self).encoded_len()
-    }
+impl TryFrom<Any<'_>> for ObjectIdentifier {
+    type Error = Error;
 
-    fn encode(&self, encoder: &mut Encoder<'_>) -> Result<()> {
-        Any::from(self).encode(encoder)
+    fn try_from(any: Any<'_>) -> Result<ObjectIdentifier> {
+        any.tag().assert_eq(Tag::ObjectIdentifier)?;
+        Ok(ObjectIdentifier::from_bytes(any.value())?)
     }
 }
 
-impl<'a> Tagged for ObjectIdentifier {
-    const TAG: Tag = Tag::ObjectIdentifier;
-}
-
 #[cfg(test)]
 mod tests {
     use super::ObjectIdentifier;
     use crate::{Decodable, Encodable, Length};
-    use core::convert::TryInto;
 
     const EXAMPLE_OID: ObjectIdentifier = ObjectIdentifier::new("1.2.840.113549");
     const EXAMPLE_OID_BYTES: &[u8; 8] = &[0x06, 0x06, 0x2a, 0x86, 0x48, 0x86, 0xf7, 0x0d];
@@ -69,6 +79,6 @@ mod tests {
     #[test]
     fn length() {
         // Ensure an infallible `From` conversion to `Any` will never panic
-        assert!(ObjectIdentifier::MAX_LENGTH <= Length::MAX.try_into().unwrap());
+        assert!(ObjectIdentifier::MAX_SIZE <= Length::MAX.try_into().unwrap());
     }
 }
