@@ -1,9 +1,10 @@
 //! ASN.1 `PrintableString` support.
 
 use crate::{
-    asn1::Any, str_slice::StrSlice, Encodable, Encoder, Error, Length, Result, Tag, Tagged,
+    asn1::Any, ByteSlice, DecodeValue, Decoder, EncodeValue, Encoder, Error, FixedTag, Length,
+    OrdIsValueOrd, Result, StrSlice, Tag,
 };
-use core::{convert::TryFrom, fmt, str};
+use core::{fmt, str};
 
 /// ASN.1 `PrintableString` type.
 ///
@@ -105,6 +106,28 @@ impl AsRef<[u8]> for PrintableString<'_> {
     }
 }
 
+impl<'a> DecodeValue<'a> for PrintableString<'a> {
+    fn decode_value(decoder: &mut Decoder<'a>, length: Length) -> Result<Self> {
+        Self::new(ByteSlice::decode_value(decoder, length)?.as_bytes())
+    }
+}
+
+impl<'a> EncodeValue for PrintableString<'a> {
+    fn value_len(&self) -> Result<Length> {
+        self.inner.value_len()
+    }
+
+    fn encode_value(&self, encoder: &mut Encoder<'_>) -> Result<()> {
+        self.inner.encode_value(encoder)
+    }
+}
+
+impl FixedTag for PrintableString<'_> {
+    const TAG: Tag = Tag::PrintableString;
+}
+
+impl OrdIsValueOrd for PrintableString<'_> {}
+
 impl<'a> From<&PrintableString<'a>> for PrintableString<'a> {
     fn from(value: &PrintableString<'a>) -> PrintableString<'a> {
         *value
@@ -115,8 +138,7 @@ impl<'a> TryFrom<Any<'a>> for PrintableString<'a> {
     type Error = Error;
 
     fn try_from(any: Any<'a>) -> Result<PrintableString<'a>> {
-        any.tag().assert_eq(Tag::PrintableString)?;
-        Self::new(any.as_bytes())
+        any.decode_into()
     }
 }
 
@@ -132,20 +154,6 @@ impl<'a> From<PrintableString<'a>> for &'a [u8] {
     }
 }
 
-impl<'a> Encodable for PrintableString<'a> {
-    fn encoded_len(&self) -> Result<Length> {
-        Any::from(*self).encoded_len()
-    }
-
-    fn encode(&self, encoder: &mut Encoder<'_>) -> Result<()> {
-        Any::from(*self).encode(encoder)
-    }
-}
-
-impl<'a> Tagged for PrintableString<'a> {
-    const TAG: Tag = Tag::PrintableString;
-}
-
 impl<'a> fmt::Display for PrintableString<'a> {
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         f.write_str(self.as_str())
