@@ -1,7 +1,9 @@
 //! DER encoder.
 
-use crate::{asn1::*, message, Encodable, Error, ErrorKind, Header, Length, Result, Tag};
-use core::convert::{TryFrom, TryInto};
+use crate::{
+    asn1::*, Encodable, EncodeValue, Error, ErrorKind, Header, Length, Result, Tag, TagMode,
+    TagNumber, Tagged,
+};
 
 /// DER encoder.
 #[derive(Debug)]
@@ -61,12 +63,14 @@ impl<'a> Encoder<'a> {
         let range = ..usize::try_from(self.position)?;
 
         match self.bytes {
-            Some(bytes) => bytes.get(range).ok_or_else(|| ErrorKind::Truncated.at(pos)),
+            Some(bytes) => bytes
+                .get(range)
+                .ok_or_else(|| ErrorKind::Overlength.at(pos)),
             None => Err(ErrorKind::Failed.at(pos)),
         }
     }
 
-    /// Encode the provided value as an ASN.1 `BIT STRING`
+    /// Encode the provided value as an ASN.1 `BIT STRING`.
     pub fn bit_string(&mut self, value: impl TryInto<BitString<'a>>) -> Result<()> {
         value
             .try_into()
@@ -74,6 +78,24 @@ impl<'a> Encoder<'a> {
             .and_then(|value| self.encode(&value))
     }
 
+    /// Encode a `CONTEXT-SPECIFIC` field with `EXPLICIT` tagging.
+    pub fn context_specific<T>(
+        &mut self,
+        tag_number: TagNumber,
+        tag_mode: TagMode,
+        value: &T,
+    ) -> Result<()>
+    where
+        T: EncodeValue + Tagged,
+    {
+        ContextSpecificRef {
+            tag_number,
+            tag_mode,
+            value,
+        }
+        .encode(self)
+    }
+
     /// Encode the provided value as an ASN.1 `GeneralizedTime`
     pub fn generalized_time(&mut self, value: impl TryInto<GeneralizedTime>) -> Result<()> {
         value
@@ -82,7 +104,7 @@ impl<'a> Encoder<'a> {
             .and_then(|value| self.encode(&value))
     }
 
-    /// Encode the provided value as an ASN.1 `IA5String`
+    /// Encode the provided value as an ASN.1 `IA5String`.
     pub fn ia5_string(&mut self, value: impl TryInto<Ia5String<'a>>) -> Result<()> {
         value
             .try_into()
@@ -90,20 +112,6 @@ impl<'a> Encoder<'a> {
             .and_then(|value| self.encode(&value))
     }
 
-    /// Encode a message with the provided [`Encodable`] fields as an
-    /// ASN.1 `SEQUENCE`.
-    pub fn message(&mut self, fields: &[&dyn Encodable]) -> Result<()> {
-        let length = message::encoded_len_inner(fields)?;
-
-        self.sequence(length, |nested_encoder| {
-            for field in fields {
-                field.encode(nested_encoder)?;
-            }
-
-            Ok(())
-        })
-    }
-
     /// Encode an ASN.1 `NULL` value.
     pub fn null(&mut self) -> Result<()> {
         self.encode(&Null)
@@ -207,7 +215,7 @@ impl<'a> Encoder<'a> {
                 *b = byte;
                 Ok(())
             }
-            None => self.error(ErrorKind::Truncated),
+            None => self.error(ErrorKind::Overlength),
         }
     }
 
@@ -232,15 +240,18 @@ impl<'a> Encoder<'a> {
 
         buffer_len
             .checked_sub(self.position.try_into()?)
-            .ok_or_else(|| ErrorKind::Truncated.at(self.position))
+            .ok_or_else(|| ErrorKind::Overlength.at(self.position))
             .and_then(TryInto::try_into)
     }
 }
 
 #[cfg(test)]
 mod tests {
+    use hex_literal::hex;
+
+    use crate::{asn1::BitString, Encodable, ErrorKind, Length, TagMode, TagNumber};
+
     use super::Encoder;
-    use crate::{Encodable, ErrorKind, Length};
 
     #[test]
     fn overlength_message() {
@@ -250,4 +261,27 @@ mod tests {
         assert_eq!(err.kind(), ErrorKind::Overlength);
         assert_eq!(err.position(), Some(Length::ZERO));
     }
+
+    #[test]
+    fn context_specific_with_implicit_field() {
+        // From RFC8410 Section 10.3:
+        // <https://datatracker.ietf.org/doc/html/rfc8410#section-10.3>
+        //
+        //    81  33:   [1] 00 19 BF 44 09 69 84 CD FE 85 41 BA C1 67 DC 3B
+        //                  96 C8 50 86 AA 30 B6 B6 CB 0C 5C 38 AD 70 31 66
+        //                  E1
+        const EXPECTED_BYTES: &[u8] =
+            &hex!("81210019BF44096984CDFE8541BAC167DC3B96C85086AA30B6B6CB0C5C38AD703166E1");
+
+        let tag_number = TagNumber::new(1);
+        let bit_string = BitString::from_bytes(&EXPECTED_BYTES[3..]).unwrap();
+
+        let mut buf = [0u8; EXPECTED_BYTES.len()];
+        let mut encoder = Encoder::new(&mut buf);
+        encoder
+            .context_specific(tag_number, TagMode::Implicit, &bit_string)
+            .unwrap();
+
+        assert_eq!(EXPECTED_BYTES, encoder.finish().unwrap());
+    }
 }
