@@ -8,6 +8,9 @@ use core::{convert::Infallible, fmt};
 #[cfg(feature = "oid")]
 use crate::asn1::ObjectIdentifier;
 
+#[cfg(feature = "pem")]
+use crate::pem;
+
 /// Result type.
 pub type Result<T> = core::result::Result<T, Error>;
 
@@ -30,6 +33,11 @@ impl Error {
         }
     }
 
+    /// If the error's `kind` is an [`ErrorKind::Incomplete`], return the `expected_len`.
+    pub fn incomplete(self) -> Option<usize> {
+        self.kind().incomplete()
+    }
+
     /// Get the [`ErrorKind`] which occurred.
     pub fn kind(self) -> ErrorKind {
         self.kind
@@ -42,7 +50,7 @@ impl Error {
 
     /// For errors occurring inside of a nested message, extend the position
     /// count by the location where the nested message occurs.
-    pub fn nested(self, nested_position: Length) -> Self {
+    pub(crate) fn nested(self, nested_position: Length) -> Self {
         // TODO(tarcieri): better handle length overflows occurring in this calculation?
         let position = (nested_position + self.position.unwrap_or_default()).ok();
 
@@ -53,6 +61,9 @@ impl Error {
     }
 }
 
+#[cfg(feature = "std")]
+impl std::error::Error for Error {}
+
 impl fmt::Display for Error {
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         write!(f, "{}", self.kind)?;
@@ -92,22 +103,41 @@ impl From<Utf8Error> for Error {
 #[cfg(feature = "oid")]
 impl From<const_oid::Error> for Error {
     fn from(_: const_oid::Error) -> Error {
-        ErrorKind::MalformedOid.into()
+        ErrorKind::OidMalformed.into()
+    }
+}
+
+#[cfg(feature = "pem")]
+impl From<pem::Error> for Error {
+    fn from(err: pem::Error) -> Error {
+        ErrorKind::Pem(err).into()
     }
 }
 
 #[cfg(feature = "std")]
-impl std::error::Error for ErrorKind {}
+impl From<std::io::Error> for Error {
+    fn from(err: std::io::Error) -> Error {
+        match err.kind() {
+            std::io::ErrorKind::NotFound => ErrorKind::FileNotFound,
+            std::io::ErrorKind::PermissionDenied => ErrorKind::PermissionDenied,
+            other => ErrorKind::Io(other),
+        }
+        .into()
+    }
+}
 
+#[cfg(feature = "time")]
+impl From<time::error::ComponentRange> for Error {
+    fn from(_: time::error::ComponentRange) -> Error {
+        ErrorKind::DateTime.into()
+    }
+}
 /// Error type.
 #[derive(Copy, Clone, Debug, Eq, PartialEq)]
 #[non_exhaustive]
 pub enum ErrorKind {
-    /// Indicates a field which is duplicated when only one is expected.
-    DuplicateField {
-        /// Tag of the duplicated field.
-        tag: Tag,
-    },
+    /// Date-and-time related errors.
+    DateTime,
 
     /// This error indicates a previous DER parsing operation resulted in
     /// an error and tainted the state of a `Decoder` or `Encoder`.
@@ -116,6 +146,31 @@ pub enum ErrorKind {
     /// subsequently resumed.
     Failed,
 
+    /// File not found error.
+    #[cfg(feature = "std")]
+    #[cfg_attr(docsrs, doc(cfg(feature = "std")))]
+    FileNotFound,
+
+    /// Message is incomplete and does not contain all of the expected data.
+    Incomplete {
+        /// Expected message length.
+        ///
+        /// Note that this length represents a *minimum* lower bound on how
+        /// much additional data is needed to continue parsing the message.
+        ///
+        /// It's possible upon subsequent message parsing that the parser will
+        /// discover even more data is needed.
+        expected_len: Length,
+
+        /// Actual length of the message buffer currently being processed.
+        actual_len: Length,
+    },
+
+    /// I/O errors.
+    #[cfg(feature = "std")]
+    #[cfg_attr(docsrs, doc(cfg(feature = "std")))]
+    Io(std::io::ErrorKind),
+
     /// Incorrect length for a given field.
     Length {
         /// Tag of the value being decoded.
@@ -128,49 +183,8 @@ pub enum ErrorKind {
         tag: Tag,
     },
 
-    /// Malformed OID
-    MalformedOid,
-
-    /// Integer overflow occurred (library bug!).
-    Overflow,
-
-    /// Message is longer than this library's internal limits support.
-    Overlength,
-
-    /// Undecoded trailing data at end of message.
-    TrailingData {
-        /// Length of the decoded data.
-        decoded: Length,
-
-        /// Total length of the remaining data left in the buffer.
-        remaining: Length,
-    },
-
-    /// Unexpected end-of-message/nested field when decoding.
-    Truncated,
-
-    /// Encoded message is shorter than the expected length.
-    ///
-    /// (i.e. an `Encodable` impl on a particular type has a buggy `encoded_len`)
-    Underlength {
-        /// Expected length
-        expected: Length,
-
-        /// Actual length
-        actual: Length,
-    },
-
-    /// Unexpected tag.
-    UnexpectedTag {
-        /// Tag the decoder was expecting (if there is a single such tag).
-        ///
-        /// `None` if multiple tags are expected/allowed, but the `actual` tag
-        /// does not match any of them.
-        expected: Option<Tag>,
-
-        /// Actual tag encountered in the message.
-        actual: Tag,
-    },
+    /// OID is improperly encoded.
+    OidMalformed,
 
     /// Unknown OID.
     ///
@@ -182,17 +196,67 @@ pub enum ErrorKind {
     /// contribute upstream support for algorithms they care about).
     #[cfg(feature = "oid")]
     #[cfg_attr(docsrs, doc(cfg(feature = "oid")))]
-    UnknownOid {
+    OidUnknown {
         /// OID value that was unrecognized by a parser for a DER-based format.
         oid: ObjectIdentifier,
     },
 
+    /// `SET` ordering error: items not in canonical order.
+    SetOrdering,
+
+    /// Integer overflow occurred (library bug!).
+    Overflow,
+
+    /// Message is longer than this library's internal limits support.
+    Overlength,
+
+    /// PEM encoding errors.
+    #[cfg(feature = "pem")]
+    #[cfg_attr(docsrs, doc(cfg(feature = "pem")))]
+    Pem(pem::Error),
+
+    /// Permission denied reading file.
+    #[cfg(feature = "std")]
+    #[cfg_attr(docsrs, doc(cfg(feature = "std")))]
+    PermissionDenied,
+
+    /// Unknown tag mode.
+    TagModeUnknown,
+
+    /// Invalid tag number.
+    ///
+    /// The "tag number" is the lower 5-bits of a tag's octet.
+    /// This error occurs in the case that all 5-bits are set to `1`,
+    /// which indicates a multi-byte tag which is unsupported by this library.
+    TagNumberInvalid,
+
+    /// Unexpected tag.
+    TagUnexpected {
+        /// Tag the decoder was expecting (if there is a single such tag).
+        ///
+        /// `None` if multiple tags are expected/allowed, but the `actual` tag
+        /// does not match any of them.
+        expected: Option<Tag>,
+
+        /// Actual tag encountered in the message.
+        actual: Tag,
+    },
+
     /// Unknown/unsupported tag.
-    UnknownTag {
+    TagUnknown {
         /// Raw byte value of the tag.
         byte: u8,
     },
 
+    /// Undecoded trailing data at end of message.
+    TrailingData {
+        /// Length of the decoded data.
+        decoded: Length,
+
+        /// Total length of the remaining data left in the buffer.
+        remaining: Length,
+    },
+
     /// UTF-8 errors.
     Utf8(Utf8Error),
 
@@ -209,34 +273,52 @@ impl ErrorKind {
     pub fn at(self, position: Length) -> Error {
         Error::new(self, position)
     }
+
+    /// If this is an [`ErrorKind::Incomplete`], return the `expected_len`.
+    pub fn incomplete(self) -> Option<usize> {
+        match self {
+            ErrorKind::Incomplete { expected_len, .. } => usize::try_from(expected_len).ok(),
+            _ => None,
+        }
+    }
 }
 
 impl fmt::Display for ErrorKind {
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         match self {
-            ErrorKind::DuplicateField { tag } => write!(f, "duplicate field for {}", tag),
+            ErrorKind::DateTime => write!(f, "date/time error"),
             ErrorKind::Failed => write!(f, "operation failed"),
+            #[cfg(feature = "std")]
+            ErrorKind::FileNotFound => f.write_str("file not found"),
+            ErrorKind::Incomplete {
+                expected_len,
+                actual_len,
+            } => write!(
+                f,
+                "ASN.1 DER message is incomplete: expected {}, actual {}",
+                expected_len, actual_len
+            ),
+            #[cfg(feature = "std")]
+            ErrorKind::Io(err) => write!(f, "I/O error: {:?}", err),
             ErrorKind::Length { tag } => write!(f, "incorrect length for {}", tag),
             ErrorKind::Noncanonical { tag } => {
                 write!(f, "ASN.1 {} not canonically encoded as DER", tag)
             }
-            ErrorKind::MalformedOid => write!(f, "malformed OID"),
-            ErrorKind::Overflow => write!(f, "integer overflow"),
-            ErrorKind::Overlength => write!(f, "DER message is too long"),
-            ErrorKind::TrailingData { decoded, remaining } => {
-                write!(
-                    f,
-                    "trailing data at end of DER message: decoded {} bytes, {} bytes remaining",
-                    decoded, remaining
-                )
+            ErrorKind::OidMalformed => write!(f, "malformed OID"),
+            #[cfg(feature = "oid")]
+            ErrorKind::OidUnknown { oid } => {
+                write!(f, "unknown/unsupported OID: {}", oid)
             }
-            ErrorKind::Truncated => write!(f, "DER message is truncated"),
-            ErrorKind::Underlength { expected, actual } => write!(
-                f,
-                "DER message too short: expected {}, got {}",
-                expected, actual
-            ),
-            ErrorKind::UnexpectedTag { expected, actual } => {
+            ErrorKind::SetOrdering => write!(f, "ordering error"),
+            ErrorKind::Overflow => write!(f, "integer overflow"),
+            ErrorKind::Overlength => write!(f, "ASN.1 DER message is too long"),
+            #[cfg(feature = "pem")]
+            ErrorKind::Pem(e) => write!(f, "PEM error: {}", e),
+            #[cfg(feature = "std")]
+            ErrorKind::PermissionDenied => f.write_str("permission denied"),
+            ErrorKind::TagModeUnknown => write!(f, "unknown tag mode"),
+            ErrorKind::TagNumberInvalid => write!(f, "invalid tag number"),
+            ErrorKind::TagUnexpected { expected, actual } => {
                 write!(f, "unexpected ASN.1 DER tag: ")?;
 
                 if let Some(tag) = expected {
@@ -245,13 +327,16 @@ impl fmt::Display for ErrorKind {
 
                 write!(f, "got {}", actual)
             }
-            #[cfg(feature = "oid")]
-            ErrorKind::UnknownOid { oid } => {
-                write!(f, "unknown/unsupported OID: {}", oid)
-            }
-            ErrorKind::UnknownTag { byte } => {
+            ErrorKind::TagUnknown { byte } => {
                 write!(f, "unknown/unsupported ASN.1 DER tag: 0x{:02x}", byte)
             }
+            ErrorKind::TrailingData { decoded, remaining } => {
+                write!(
+                    f,
+                    "trailing data at end of DER message: decoded {} bytes, {} bytes remaining",
+                    decoded, remaining
+                )
+            }
             ErrorKind::Utf8(e) => write!(f, "{}", e),
             ErrorKind::Value { tag } => write!(f, "malformed ASN.1 DER value for {}", tag),
         }
